
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
__kernel void Reduction_InterleavedAddressing(__global uint* array, uint stride) 
{
	int GID = get_global_id(0);
	int pos1 = GID*2*stride;
	int pos2 = pos1+stride;

	array[pos1] = array[pos1] + array[pos2];
	array[pos2] =  0;

}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
__kernel void Reduction_SequentialAddressing(__global uint* array, uint stride) 
{
	int GID = get_global_id(0);
	int pos2 = GID + stride;

	array[GID] = array[GID] + array[pos2];

}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
__kernel void Reduction_Decomp(const __global uint* inArray, __global uint* outArray, uint N, __local uint* localBlock)
{
	int ls = get_local_size(0);
	int gs = get_global_size(0);

	int gid = get_global_id(0);
	int lid = get_local_id(0);

	int groupID = get_group_id(0);

	localBlock[lid] = inArray[gid] + inArray[gid + gs];

	barrier(CLK_LOCAL_MEM_FENCE);

	for(uint stride = ls / 2; stride > 0; stride /= 2){
		if(lid < stride){
			localBlock[lid] = localBlock[lid] + localBlock[lid + stride];
		}
		barrier(CLK_LOCAL_MEM_FENCE);
	}

	if(lid == 0)
		outArray[groupID] = localBlock[0]; 
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
__kernel void Reduction_DecompUnroll(const __global uint* inArray, __global uint* outArray, uint N, __local uint* localBlock)
{
		int ls = get_local_size(0);
	int gs = get_global_size(0);

	int gid = get_global_id(0);
	int lid = get_local_id(0);

	int groupID = get_group_id(0);

	localBlock[lid] = inArray[gid] + inArray[gid + gs];

	barrier(CLK_LOCAL_MEM_FENCE);

	for(uint stride = ls / 2; stride > 32; stride /= 2){
		if(lid < stride){
			localBlock[lid] = localBlock[lid] + localBlock[lid + stride];
		}
		barrier(CLK_LOCAL_MEM_FENCE);
	}

	if (lid < 32) {
		if (ls >= 64) localBlock[lid] += localBlock[lid + 32];
		if (ls >= 32) localBlock[lid] += localBlock[lid + 16];
		if (ls >= 16) localBlock[lid] += localBlock[lid + 8];
		if (ls >= 8) localBlock[lid] += localBlock[lid + 4];
		if (ls >= 4) localBlock[lid] += localBlock[lid + 2];
	}

	if(lid == 0)
		outArray[groupID] = localBlock[0]+localBlock[1]; 
}
