


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
__kernel void Scan_Naive(const __global uint* inArray, __global uint* outArray, uint N, uint offset) 
{
	int left = get_global_id(0);
	int right = left + offset;

	int value  = inArray[left];

	if(left < offset){
		outArray[left] = value;
	}

	if (right < N){
		outArray[right] = value + inArray[right];
	}
}



// Why did we not have conflicts in the Reduction? Because of the sequential addressing (here we use interleaved => we have conflicts).

#define UNROLL
#define NUM_BANKS			32
#define NUM_BANKS_LOG		5
#define SIMD_GROUP_SIZE		32

// Bank conflicts
#define AVOID_BANK_CONFLICTS
#ifdef AVOID_BANK_CONFLICTS
	// TO DO: define your conflict-free macro here
	#define OFFSET(A) ((A) + (A)/NUM_BANKS)
#else
	#define OFFSET(A) (A)
#endif

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
__kernel void Scan_WorkEfficient(__global uint* array, __global uint* higherLevelArray, __local uint* localBlock) 
{
	int ls = get_local_size(0);
	int ls2 = ls * 2;

	int gid2 = get_global_id(0) *2;
	int lid = get_local_id(0);
	int lid2 = lid *2;

	//read to local
	uint inLeft, inRight;;
	localBlock[OFFSET(lid2)] = inLeft = array[gid2];
	localBlock[OFFSET(lid2+1)] = inRight = array[gid2+1];
	barrier(CLK_LOCAL_MEM_FENCE);

	//Upsweep
	for (uint stride = 1; stride <= ls; stride *= 2) {
		
		uint right = (ls2 - 1) - lid * stride * 2;
		uint left = right - stride;

		if (right < ls2 && left < ls2) {
			localBlock[OFFSET(right)] += localBlock[OFFSET(left)];
		}

		barrier(CLK_LOCAL_MEM_FENCE);
	}

	//letzten auf 0 setzen 
	if(lid == 0)
		localBlock[OFFSET(ls2 - 1)] = 0;

	barrier(CLK_LOCAL_MEM_FENCE);

	//Downsweep
	for (uint stride = ls; stride >= 1; stride /= 2) {
		
		uint right = (ls2 - 1) - lid * stride * 2;
		uint left = right - stride;

		if (right < ls2 && left < ls2) {
			uint vright = localBlock[OFFSET(right)];
			uint vleft = localBlock[OFFSET(left)];
			localBlock[OFFSET(left)] = vright;
			localBlock[OFFSET(right)] = vleft + vright;
		}
		
		barrier(CLK_LOCAL_MEM_FENCE);
	}

	//write back
	array[gid2] = localBlock[OFFSET(lid2)] + inLeft;
	uint last = localBlock[OFFSET(lid2 + 1)] + inRight;
	array[gid2 + 1] = last;

	if (lid + 1 == ls) {
		higherLevelArray[get_group_id(0)] = last;
	}
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
__kernel void Scan_WorkEfficientAdd(__global uint* higherLevelArray, __global uint* array, __local uint* localBlock) 
{
	int ls2 = get_local_size(0) * 2;
	int gs2 = get_global_size(0) * 2;

	int lid = get_local_id(0);
	int lid2 = lid *2;

	uint id = (get_group_id(0) + 1) * ls2 + lid2;

	uint add = higherLevelArray[get_group_id(0)];

	if (id < gs2) {
		array[id] += add;
	}
	if (id + 1 < gs2) {
		array[id + 1] += add;
	}

	// TO DO: Kernel implementation (large arrays)
	// Kernel that should add the group PPS to the local PPS (Figure 14)
}